# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Collection of argparse actions for django-ca management commands."""

import abc
import argparse
import getpass
import typing
from datetime import timedelta
from typing import Any, List, Optional, Type

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

from django_ca import ca_settings, constants
from django_ca.constants import EXTENSION_DEFAULT_CRITICAL, KEY_USAGE_NAMES, ReasonFlags
from django_ca.models import Certificate, CertificateAuthority
from django_ca.typehints import AllowedHashTypes, AlternativeNameExtensionType
from django_ca.utils import is_power2, parse_encoding, parse_general_name, x509_name

ActionType = typing.TypeVar("ActionType")  # pylint: disable=invalid-name
ParseType = typing.TypeVar("ParseType")  # pylint: disable=invalid-name
ExtensionType = typing.TypeVar("ExtensionType", bound=x509.ExtensionType)  # pylint: disable=invalid-name


def general_name_type(value: str) -> x509.GeneralName:
    """Wrapper function to parse_general_name() that sets a name."""
    return parse_general_name(value)


general_name_type.__name__ = "general name"


class SingleValueAction(argparse.Action, typing.Generic[ParseType, ActionType], metaclass=abc.ABCMeta):
    """Abstract/generic base class for arguments that take a single value.

    The main purpose of this class is to improve type hinting.
    """

    type: Type[ActionType]

    @abc.abstractmethod
    def parse_value(self, value: ParseType) -> ActionType:
        """Parse the value passed to the command line. Implementing classes must implement this method.

        Parameters
        ----------

        value : str
            The value passed by the command line.
        """
        raise NotImplementedError

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: ParseType,
        option_string: Optional[str] = None,
    ) -> None:
        setattr(namespace, self.dest, self.parse_value(values))


class AlgorithmAction(SingleValueAction[str, AllowedHashTypes]):
    """Action for giving an algorithm.

    >>> parser.add_argument('--algorithm', action=AlgorithmAction)  # doctest: +ELLIPSIS
    AlgorithmAction(...)
    >>> parser.parse_args(['--algorithm', 'SHA-256'])  # doctest: +ELLIPSIS
    Namespace(algorithm=<cryptography.hazmat.primitives.hashes.SHA256 object at ...>)
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("choices", sorted(tuple(constants.HASH_ALGORITHM_TYPES)))
        kwargs.setdefault("metavar", "{SHA-512,SHA-256,...}")
        super().__init__(**kwargs)

    def parse_value(self, value: str) -> AllowedHashTypes:
        """Parse the value for this action."""
        # NOTE: A KeyError is ruled out by the choices argument set in the constructor.
        return constants.HASH_ALGORITHM_TYPES[value]()


class CertificateAction(SingleValueAction[str, Certificate]):
    """Action for naming a certificate."""

    def __init__(self, allow_revoked: bool = False, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.allow_revoked = allow_revoked

    def parse_value(self, value: str) -> Certificate:
        """Parse the value for this action."""
        queryset = Certificate.objects.all()
        if self.allow_revoked is False:
            queryset = queryset.filter(revoked=False)

        try:
            return queryset.get_by_serial_or_cn(value)
        except Certificate.DoesNotExist as ex:
            raise argparse.ArgumentError(self, f"{value}: Certificate not found.") from ex
        except Certificate.MultipleObjectsReturned as ex:
            raise argparse.ArgumentError(self, f"{value}: Multiple certificates match.") from ex


class CertificateAuthorityAction(SingleValueAction[str, CertificateAuthority]):
    """Action for naming a certificate authority."""

    def __init__(self, allow_disabled: bool = False, allow_unusable: bool = False, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.allow_disabled = allow_disabled
        self.allow_unusable = allow_unusable

    def parse_value(self, value: str) -> CertificateAuthority:
        """Parse the value for this action."""
        qs = CertificateAuthority.objects.all()
        if self.allow_disabled is False:
            qs = qs.enabled()

        try:
            ca = qs.get_by_serial_or_cn(value)
        except CertificateAuthority.DoesNotExist as ex:
            raise argparse.ArgumentError(self, f"{value}: Certificate authority not found.") from ex
        except CertificateAuthority.MultipleObjectsReturned as ex:
            raise argparse.ArgumentError(self, f"{value}: Multiple Certificate authorities match.") from ex

        # verify that the private key exists
        if not self.allow_unusable and not ca.key_exists:
            raise argparse.ArgumentError(self, f"{ca}: {ca.private_key_path}: Private key does not exist.")

        return ca


class ExpiresAction(SingleValueAction[str, timedelta]):
    """Action for passing a timedelta in days.

    NOTE: str(timedelta) is different in python 3.6, so only outputting days here

    >>> parser.add_argument('--expires', action=ExpiresAction)  # doctest: +ELLIPSIS
    ExpiresAction(...)
    >>> parser.parse_args(['--expires', '3']).expires.days
    3
    """

    def parse_value(self, value: str) -> timedelta:
        """Parse the value for this action."""
        # NOTE: Making this a member of ExpiresAction causes an infinite loop for some reason
        try:
            days = int(value)
        except ValueError as ex:
            raise argparse.ArgumentError(self, f"{value}: Value must be an integer.") from ex
        if days <= 0:
            raise argparse.ArgumentError(self, f"{value}: Value must not be negative.")

        return timedelta(days=days)


class FormatAction(SingleValueAction[str, Encoding]):
    """Action for giving an encoding (DER/PEM).

    >>> parser.add_argument('--format', action=FormatAction)  # doctest: +ELLIPSIS
    FormatAction(...)
    >>> parser.parse_args(['--format', 'DER'])
    Namespace(format=<Encoding.DER: 'DER'>)
    """

    def parse_value(self, value: str) -> Encoding:
        """Parse the value for this action."""
        try:
            return parse_encoding(value)
        except ValueError as e:
            raise argparse.ArgumentError(self, str(e))


class EllipticCurveAction(SingleValueAction[str, ec.EllipticCurve]):
    """Action to parse an elliptic curve value.

    >>> parser.add_argument('--elliptic-curve', action=EllipticCurveAction)  # doctest: +ELLIPSIS
    EllipticCurveAction(...)
    >>> parser.parse_args(['--elliptic-curve', 'secp256r1'])  # doctest: +ELLIPSIS
    Namespace(elliptic_curve=<cryptography.hazmat.primitives.asymmetric.ec.SECP256R1 object at ...>)
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("choices", sorted(tuple(constants.ELLIPTIC_CURVE_TYPES)))
        kwargs.setdefault("metavar", "{secp256r1,secp384r1,secp521r1,...}")
        super().__init__(**kwargs)

    def parse_value(self, value: str) -> ec.EllipticCurve:
        """Parse the value for this action."""
        # NOTE: A KeyError is ruled out by the choices argument set in the constructor.
        return constants.ELLIPTIC_CURVE_TYPES[value]()


class IntegerRangeAction(SingleValueAction[int, int]):
    """An int action with an optional min/max value.

    >>> parser.add_argument('--size', action=IntegerRangeAction, min=0, max=10)  # doctest: +ELLIPSIS
    IntegerRangeAction(...)
    >>> parser.parse_args(['--size', '5'])
    Namespace(size=5)

    Parameters
    ----------

    min: int, Optional
        The optional minimum value.
    max: int, Optional
        The optional maximum value.
    """

    def __init__(self, **kwargs: Any) -> None:
        self.min = kwargs.pop("min", None)
        self.max = kwargs.pop("max", None)
        kwargs["type"] = int  # so parse_value() will receive an int
        kwargs.setdefault("metavar", "INT")
        super().__init__(**kwargs)

    def parse_value(self, value: int) -> int:
        if self.min is not None and self.min > value:
            raise argparse.ArgumentError(self, f"{self.metavar} must be equal or greater then {self.min}.")
        if self.max is not None and self.max < value:
            raise argparse.ArgumentError(self, f"{self.metavar} must be equal or smaller then {self.max}.")
        return value


class KeySizeAction(SingleValueAction[str, int]):
    """Action for adding a keysize, an integer that must be a power of two (2048, 4096, ...).

    >>> parser.add_argument('--size', action=KeySizeAction)  # doctest: +ELLIPSIS
    KeySizeAction(...)
    >>> parser.parse_args(['--size', '4096'])
    Namespace(size=4096)
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("metavar", "{2048,4096,8192,...}")
        super().__init__(**kwargs)

    def parse_value(self, value: str) -> int:
        """Parse the value for this action."""
        try:
            key_size = int(value)
        except ValueError as ex:
            raise argparse.ArgumentError(self, f"{value}: Must be an integer.") from ex

        if not is_power2(key_size):
            raise argparse.ArgumentError(self, f"{key_size}: Must be a power of two (2048, 4096, ...).")

        if key_size < ca_settings.CA_MIN_KEY_SIZE:
            raise argparse.ArgumentError(
                self, f"{key_size}: Must be at least {ca_settings.CA_MIN_KEY_SIZE} bits."
            )
        return key_size


class MultipleURLAction(argparse.Action):
    """Action for multiple URLs.

    >>> parser.add_argument('--url', action=MultipleURLAction)  # doctest: +ELLIPSIS
    MultipleURLAction(...)
    >>> parser.parse_args(['--url', 'https://example.com', '--url', 'https://example.net'])
    Namespace(url=['https://example.com', 'https://example.net'])
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("default", [])
        kwargs.setdefault("metavar", "URL")
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        validator = URLValidator()
        try:
            validator(values)
        except ValidationError as ex:
            raise argparse.ArgumentError(self, f"{values}: Not a valid URL.") from ex

        getattr(namespace, self.dest).append(values)


class PasswordAction(argparse.Action):
    """Action for adding a password argument.

    If the cli does not pass an argument value, the action prompt the user for a password.

    >>> parser.add_argument('--password', action=PasswordAction)  # doctest: +ELLIPSIS
    PasswordAction(...)
    >>> parser.parse_args(['--password', 'secret'])
    Namespace(password=b'secret')
    """

    def __init__(self, prompt: str = "Password: ", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.prompt = prompt

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Optional[str],
        option_string: Optional[str] = None,
    ) -> None:
        if values is None:
            values = getpass.getpass(prompt=self.prompt)

        setattr(namespace, self.dest, values.encode("utf-8"))


class CertificationPracticeStatementAction(argparse.Action):
    """Add a Certification Practice Statement to a previously added Certificate Policy.

    This action works in tandem with :py:class:`~django_ca.management.actions.PolicyIdentifierAction`, and has
    to be called after that action to add a `policy_qualifier` to it. The `dest` arg to this action must
    match the destination of the `PolicyIdentifierAction`.

    The action verifies that the given value is a URI.

    >>> parser.add_argument('--pi', action=PolicyIdentifierAction)  # doctest: +ELLIPSIS
    PolicyIdentifierAction(...)
    >>> parser.add_argument(
    ...     '--cps', action=CertificationPracticeStatementAction, dest="pi"
    ... )  # doctest: +ELLIPSIS
    CertificationPracticeStatementAction(...)
    >>> parser.parse_args(['--pi', '1.2.3', '--cps', 'https://example.com/cps']).pi[0].policy_qualifiers
    ['https://example.com/cps']
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs["metavar"] = "URL"
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        certificate_policies = getattr(namespace, self.dest)
        # Make sure that --policy-identifier was called before
        if certificate_policies is None:
            raise argparse.ArgumentError(self, "Must be preceded by --policy-identifier.")

        # RFC 5280, section 4.2.1.4 mandates that CPS must be in the form of a URI.
        validator = URLValidator()
        try:
            validator(values)
        except ValidationError as ex:
            raise argparse.ArgumentError(self, f"{values}: Not a valid URL.") from ex

        certificate_policies[-1].policy_qualifiers.append(values)


class PolicyIdentifierAction(argparse.Action):
    """Action to add a Certificate Policies extension.

    This action adds a :py:class:`cg:~cryptography.x509.CertificatePolicies` instance to the namespace. A
    :py:class:`cg:~cryptography.x509.PolicyInformation` with the given OID as `policy_identifier` will be
    added to it. Policies given in previous iterations of this argument will be prepended.

    The `allow_any_policy` argument allows adding the ``anyPolicy`` (OID "2.5.29.32.0") policy can be added.
    This is the case for certificate authorities.

    >>> parser.add_argument('--pi', action=PolicyIdentifierAction)  # doctest: +ELLIPSIS
    PolicyIdentifierAction(...)
    >>> parser.parse_args(['--pi', '1.2.3']).pi  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    <CertificatePolicies([<PolicyInformation(policy_identifier=<ObjectIdentifier(oid=1.2.3, name=...)>,
            policy_qualifiers=[])>])>
    >>> parser.parse_args(['--pi', '2.5.29', '--pi', '1.2.3']).pi  # doctest: +ELLIPSIS  +NORMALIZE_WHITESPACE
    <CertificatePolicies([<PolicyInformation(policy_identifier=<ObjectIdentifier(oid=2.5.29, name=...)>,
            policy_qualifiers=[])>,
        <PolicyInformation(policy_identifier=<ObjectIdentifier(oid=1.2.3, name=...)>,
            policy_qualifiers=[])>])>
    """

    def __init__(self, **kwargs: Any) -> None:
        self.allow_any_policy = kwargs.pop("allow_any_policy", False)
        kwargs["metavar"] = "OID"
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        if values == "anyPolicy":
            values = "2.5.29.32.0"

        if self.allow_any_policy is False and values == "2.5.29.32.0":
            raise argparse.ArgumentError(self, "anyPolicy is not allowed in this context.")

        try:
            oid = x509.ObjectIdentifier(values)
        except ValueError as ex:
            raise argparse.ArgumentError(self, f"invalid ObjectIdentifier value: '{values}'") from ex

        policy = x509.PolicyInformation(policy_identifier=oid, policy_qualifiers=[])

        if certificate_policies := getattr(namespace, self.dest):
            policies = list(certificate_policies) + [policy]
        else:
            policies = [policy]

        setattr(namespace, self.dest, x509.CertificatePolicies(policies=policies))


class ReasonAction(SingleValueAction[str, ReasonFlags]):
    """Action to select a revocation reason.

    >>> parser.add_argument('--reason', action=ReasonAction)  # doctest: +ELLIPSIS
    ReasonAction(...)
    >>> parser.parse_args(['--reason', 'key_compromise'])
    Namespace(reason=<ReasonFlags.key_compromise: 'keyCompromise'>)
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs["choices"] = sorted([r.name for r in ReasonFlags])
        kwargs.setdefault("default", ReasonFlags.unspecified)
        super().__init__(**kwargs)

    def parse_value(self, value: str) -> ReasonFlags:
        """Parse the value for this action."""
        # NOTE: set of choices already assures that value is a valid ReasonFlag
        return ReasonFlags[value]


class NameAction(SingleValueAction[str, x509.Name]):
    """Action to parse a string into a :py:class:`cg:~cryptography.x509.Name`.

    Note that this action does *not* take care of sorting the subject in any way.

    >>> parser.add_argument('--name', action=NameAction)  # doctest: +ELLIPSIS
    NameAction(...)
    >>> parser.parse_args(["--name", "/CN=example.com"])
    Namespace(name=<Name(CN=example.com)>)
    """

    def parse_value(self, value: str) -> x509.Name:
        try:
            return x509_name(value)
        except ValueError as e:
            raise argparse.ArgumentError(self, str(e))


class URLAction(SingleValueAction[str, str]):
    """Action to pass a single valid URL.

    >>> parser.add_argument('--url', action=URLAction)  # doctest: +ELLIPSIS
    URLAction(...)
    >>> parser.parse_args(['--url', 'https://example.com'])
    Namespace(url='https://example.com')
    """

    def parse_value(self, value: str) -> str:
        """Parse the value for this action."""
        validator = URLValidator()
        try:
            validator(value)
        except ValidationError as ex:
            raise argparse.ArgumentError(self, f"{value}: Not a valid URL.") from ex

        return value


class UserNoticeAction(argparse.Action):
    """Add a User Notice to a previously added Certificate Policy.

    This action works in tandem with :py:class:`~django_ca.management.actions.PolicyIdentifierAction`, and has
    to be called after that action to add a `policy_qualifier` to it. The `dest` arg to this action must
    match the destination of the `PolicyIdentifierAction`.

    The action verifies that the given value is no longer then 200 characters (RFC 5280, section 4.2.1.4).

    >>> parser.add_argument('--pi', action=PolicyIdentifierAction)  # doctest: +ELLIPSIS
    PolicyIdentifierAction(...)
    >>> parser.add_argument('--notice', action=UserNoticeAction, dest="pi")  # doctest: +ELLIPSIS
    UserNoticeAction(...)
    >>> parser.parse_args(['--pi', '1.2.3', '--notice', 'example text']).pi[0].policy_qualifiers
    [<UserNotice(notice_reference=None, explicit_text='example text')>]
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs["metavar"] = "TEXT"
        kwargs.setdefault("dest", "certificate_policies")
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        certificate_policies = getattr(namespace, self.dest)
        # Make sure that --policy-identifier was called before
        if certificate_policies is None:
            raise argparse.ArgumentError(self, "Must be preceded by --policy-identifier.")

        # RFC 5280, section 4.2.1.4 mandates that CPS must be in the form of a URI.
        if len(values) > 200:
            raise argparse.ArgumentError(self, f"{self.metavar} must not be longer then 200 characters.")

        user_notice = x509.UserNotice(notice_reference=None, explicit_text=values)

        certificate_policies[-1].policy_qualifiers.append(user_notice)


##########################
# x509 extension actions #
##########################


class CryptographyExtensionAction(argparse.Action, typing.Generic[ExtensionType], metaclass=abc.ABCMeta):
    """Base class for actions that return a cryptography ExtensionType instance."""

    extension_type: Type[ExtensionType]


class AlternativeNameLegacyAction(CryptographyExtensionAction[AlternativeNameExtensionType]):
    """Action for AlternativeName extensions."""

    def __init__(self, extension_type: Type[AlternativeNameExtensionType], **kwargs: Any) -> None:
        self.extension_type = extension_type
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        ext = getattr(namespace, self.dest)
        if ext is None:
            names = []
        else:
            names = list(ext.value)

        names.append(parse_general_name(values))
        extension_type = self.extension_type(general_names=names)
        critical = EXTENSION_DEFAULT_CRITICAL[self.extension_type.oid]
        extension = x509.Extension(oid=self.extension_type.oid, critical=critical, value=extension_type)

        setattr(namespace, self.dest, extension)


class AlternativeNameAction(CryptographyExtensionAction[AlternativeNameExtensionType]):
    """Action for AlternativeName extensions.

    >>> parser.add_argument(
    ...     '--subject-alternative-name',
    ...     action=AlternativeNameAction,
    ...     extension_type=x509.SubjectAlternativeName
    ... )  # doctest: +ELLIPSIS
    AlternativeNameAction(...)
    >>> args = parser.parse_args(['--subject-alternative-name', 'https://example.com'])
    >>> args.subject_alternative_name  # doctest: +NORMALIZE_WHITESPACE
    <SubjectAlternativeName(<GeneralNames([<UniformResourceIdentifier(value='https://example.com')>])>)>
    """

    def __init__(self, extension_type: Type[AlternativeNameExtensionType], **kwargs: Any) -> None:
        self.extension_type = extension_type
        kwargs["metavar"] = "NAME"
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        alternative_names = getattr(namespace, self.dest)
        if alternative_names is None:
            names = []
        else:
            names = list(alternative_names)

        names.append(parse_general_name(values))
        extension_type = self.extension_type(general_names=names)

        setattr(namespace, self.dest, extension_type)


class AuthorityInformationAccessAction(CryptographyExtensionAction[x509.AuthorityInformationAccess]):
    """Action for parsing an AuthorityInformationAccess extension.

    This extension has a required parameter `access_method`, which should be the OID that it creates. You
    would usually add two arguments using this action, one for OCSP responders and one for CA issuers, and
    match the `dest` argument to produce one extension:

    >>> from cryptography.x509.oid import AuthorityInformationAccessOID
    >>> parser.add_argument(
    ...     "--ocsp-responder",
    ...     action=AuthorityInformationAccessAction,
    ...     access_method=AuthorityInformationAccessOID.OCSP,
    ...     dest="extension"
    ... )  # doctest: +ELLIPSIS
    AuthorityInformationAccessAction(...)
    >>> parser.add_argument(
    ...     "--ca-issuer",
    ...     action=AuthorityInformationAccessAction,
    ...     access_method=AuthorityInformationAccessOID.CA_ISSUERS,
    ...     dest="extension"
    ... )  # doctest: +ELLIPSIS
    AuthorityInformationAccessAction(...)
    >>> parser.parse_args(
    ...     ["--ocsp-responder", "http://ocsp", "--ca-issuer", "http://issuer"]
    ... )  # doctest: +ELLIPSIS
    Namespace(extension=<AuthorityInformationAccess(...)
    """

    def __init__(self, access_method: x509.ObjectIdentifier, **kwargs: Any) -> None:
        self.access_method = access_method
        kwargs["type"] = general_name_type
        kwargs["metavar"] = ("NAME",)
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: x509.GeneralName,
        option_string: Optional[str] = None,
    ) -> None:
        extension: Optional[x509.AuthorityInformationAccess] = getattr(namespace, self.dest)
        access_description = x509.AccessDescription(access_method=self.access_method, access_location=values)

        if extension is None:
            access_descriptions = [access_description]
        else:
            access_descriptions = list(extension) + [access_description]

        # Finally sort by OID so that we have more predictable behavior
        access_descriptions = sorted(access_descriptions, key=lambda ad: ad.access_method.dotted_string)

        extension = x509.AuthorityInformationAccess(access_descriptions)

        setattr(namespace, self.dest, extension)


class ExtendedKeyUsageAction(CryptographyExtensionAction[x509.ExtendedKeyUsage]):
    """Action for parsing an ExtendedKeyUsage extension.

    >>> parser.add_argument('--extended-key-usage', action=ExtendedKeyUsageAction)  # doctest: +ELLIPSIS
    ExtendedKeyUsageAction(...)
    >>> args = parser.parse_args(["--extended-key-usage", "serverAuth", "clientAuth"])
    >>> args.extended_key_usage  # doctest: +NORMALIZE_WHITESPACE
    <ExtendedKeyUsage([<ObjectIdentifier(oid=1.3.6.1.5.5.7.3.1, name=serverAuth)>,
                                        <ObjectIdentifier(oid=1.3.6.1.5.5.7.3.2, name=clientAuth)>])>
    """

    extension_type = x509.ExtendedKeyUsage

    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("nargs", "+")
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: List[str],
        option_string: Optional[str] = None,
    ) -> None:
        usages: List[x509.ObjectIdentifier] = []

        for value in values:
            if value in constants.EXTENDED_KEY_USAGE_OIDS:
                oid = constants.EXTENDED_KEY_USAGE_OIDS[value]
            else:
                try:
                    oid = x509.ObjectIdentifier(value)
                except ValueError as ex:
                    raise argparse.ArgumentError(
                        self, f"{value}: Not a dotted string or known Extended Key Usage."
                    ) from ex

            if oid in usages:
                raise argparse.ArgumentError(self, f"{value}: Extended Key Usage is added multiple times.")
            usages.append(oid)

        extended_key_usage = x509.ExtendedKeyUsage(usages)
        setattr(namespace, self.dest, extended_key_usage)


class KeyUsageAction(CryptographyExtensionAction[x509.KeyUsage]):
    """Action for parsing a KeyUsage extension.

    >>> parser.add_argument('--key-usage', action=KeyUsageAction)  # doctest: +ELLIPSIS
    KeyUsageAction(...)
    >>> args = parser.parse_args(['--key-usage', 'keyCertSign'])
    >>> args.key_usage  # doctest: +NORMALIZE_WHITESPACE
    <KeyUsage(digital_signature=False,
             content_commitment=False,
             key_encipherment=False,
             data_encipherment=False,
             key_agreement=False,
             key_cert_sign=True,
             crl_sign=False,
             encipher_only=False,
             decipher_only=False)>
    """

    extension_type = x509.KeyUsage

    def __init__(self, **kwargs: Any) -> None:
        # TODO: add choices once support for old comma-separated lists are removed.
        kwargs.setdefault("nargs", "+")
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: List[str],
        option_string: Optional[str] = None,
    ) -> None:
        if invalid_usages := [ku for ku in values if ku not in KEY_USAGE_NAMES.values()]:
            raise argparse.ArgumentError(
                self, f"{', '.join(sorted(set(invalid_usages)))}: Invalid key usage."
            )

        key_usages = {k: v in values for k, v in KEY_USAGE_NAMES.items()}
        try:
            extension_type = x509.KeyUsage(**key_usages)
        except ValueError as ex:
            raise argparse.ArgumentError(self, str(ex)) from ex

        setattr(namespace, self.dest, extension_type)


class TLSFeatureAction(CryptographyExtensionAction[x509.TLSFeature]):
    """Action for parsing a TLSFeature extension."""

    extension_type = x509.TLSFeature

    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("nargs", "+")
        super().__init__(**kwargs)

    def __call__(  # type: ignore[override] # argparse.Action defines much looser type for values
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: List[str],
        option_string: Optional[str] = None,
    ) -> None:
        try:
            features = [constants.TLS_FEATURE_NAMES[value] for value in values]
        except KeyError as ex:
            raise argparse.ArgumentError(self, f"Unknown TLSFeature: {ex.args[0]}")

        extension_type = x509.TLSFeature(features=features)
        setattr(namespace, self.dest, extension_type)



# Generated by Django 3.2 on 2021-04-30 11:24

from django.db import migrations
import django_ca.modelfields


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0021_auto_20210416_1557'),
    ]

    operations = [
        migrations.AddField(
            model_name='certificate',
            name='csr_tmp',
            field=django_ca.modelfields.CertificateSigningRequestField(blank=True, null=True, verbose_name='CSR'),
        ),
        migrations.AddField(
            model_name='certificate',
            name='pub_tmp',
            field=django_ca.modelfields.CertificateField(default=b'', null=True, verbose_name='Public key'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='certificateauthority',
            name='pub_tmp',
            field=django_ca.modelfields.CertificateField(default=b'', null=True, verbose_name='Public key'),
            preserve_default=False,
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Collection of Django HTTP response subclasses representing ACME responses."""

from http import HTTPStatus
from typing import Any, Optional, Type

from acme import messages

from django.http import HttpRequest, JsonResponse
from django.urls import reverse

from django_ca.acme.messages import Order
from django_ca.models import AcmeAccount


class AcmeResponse(JsonResponse):
    """Base class for all ACME responses."""


class AcmeSimpleResponse(AcmeResponse):
    """Base class for all responses returning an ACME recourses (accounts, etc.)."""

    message_cls: Type[messages.ResourceBody]

    def __init__(self, **kwargs: Any):
        super().__init__(self.message_cls(**kwargs).to_json())


class AcmeResponseAccount(AcmeResponse):
    """Response containing an ACME account."""

    def __init__(self, request: HttpRequest, account: AcmeAccount) -> None:
        contact = []
        if account.contact:
            contact = account.contact.split("\n")

        data = {
            "status": account.status,
            "contact": contact,
            "orders": request.build_absolute_uri(
                reverse(
                    "django_ca:acme-account-orders",
                    kwargs={"slug": account.slug, "serial": account.ca.serial},
                )
            ),
        }

        super().__init__(data)
        self["Location"] = account.kid


class AcmeResponseAccountCreated(AcmeResponseAccount):
    """Response when an ACME account is created."""

    status_code = HTTPStatus.CREATED  # 201


class AcmeResponseOrder(AcmeSimpleResponse):
    """An HTTP response returning an Order object."""

    message_cls = Order


class AcmeResponseOrderCreated(AcmeResponseOrder):
    """An Order response but with HTTP status code CREATED (201)."""

    status_code = HTTPStatus.CREATED  # 201


class AcmeResponseAuthorization(AcmeSimpleResponse):
    """An HTTP response for an ACME authorization."""

    message_cls = messages.Authorization


class AcmeResponseChallenge(AcmeSimpleResponse):
    """An HTTP response for an ACME challenge."""

    message_cls = messages.ChallengeBody


class AcmeResponseError(AcmeResponse):
    """Base class for all ACME error responses."""

    message = "Internal server error"
    status_code = HTTPStatus.INTERNAL_SERVER_ERROR  # 500
    type = "serverInternal"

    def __init__(self, typ: Optional[str] = None, message: str = "") -> None:
        super().__init__(
            {
                "type": f"urn:ietf:params:acme:error:{typ or self.type}",
                "status": self.status_code,
                "detail": message or self.message,
            },
            content_type="application/problem+json",
        )


class AcmeResponseMalformed(AcmeResponseError):
    """ACME response with type malformed."""

    message = "Malformed request"
    status_code = HTTPStatus.BAD_REQUEST  # 400
    type = "malformed"


class AcmeResponseBadCSR(AcmeResponseMalformed):
    """ACME response for an invalid CSR."""

    type = "badCSR"
    message = "CSR is not acceptable."


class AcmeResponseMalformedPayload(AcmeResponseMalformed):
    """A subclass of a malformed response for unparsable payloads.

    This class is reserved for cases where the payload cannot even be processed, e.g. it's missing a required
    key, or a value is invalid.
    """

    message = "Malformed payload."


class AcmeResponseUnauthorized(AcmeResponseError):
    """ACME response with type unauthorized."""

    status_code = HTTPStatus.UNAUTHORIZED  # 401
    type = "unauthorized"
    message = "You are not authorized to perform this request."


class AcmeResponseForbidden(AcmeResponseError):
    """ACME response with HTTP status code 403 (Forbidden)."""

    status_code = HTTPStatus.FORBIDDEN  # 403
    type = "forbidden"
    message = "You are forbidden from accessing this resource."


class AcmeResponseNotFound(AcmeResponseError):
    """ACME response when a requested entity cannot be found."""

    status_code = HTTPStatus.NOT_FOUND  # 404
    type = "not-found"
    message = "The requested entity could not be found."


class AcmeResponseBadNonce(AcmeResponseError):
    """

    .. seealso:: RFC 8555, section 6.5:

       "When a server rejects a request because its nonce value was unacceptable (or not present), it MUST
       provide HTTP status code 400 (Bad Request), and indicate the ACME error type
       "urn:ietf:params:acme:error:badNonce".  An error response with the "badNonce" error type MUST include a
       Replay-Nonce header field with a fresh nonce that the server will accept in a retry of the original
       query (and possibly in other requests, according to the server's nonce scoping policy)."
    """

    status_code = HTTPStatus.BAD_REQUEST  # 400
    type = "badNonce"
    message = "Bad or invalid nonce."


class AcmeResponseUnsupportedMediaType(AcmeResponseMalformed):
    """Acme response for unsupported media type."""

    status_code = HTTPStatus.UNSUPPORTED_MEDIA_TYPE
    message = "Requests must use the application/jose+json content type."



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Pydantic Schemas for the API."""

import abc
from datetime import datetime
from typing import Optional

from ninja import Field, ModelSchema, Schema

from django_ca import ca_settings, constants
from django_ca.api.extension_schemas import DATETIME_EXAMPLE, ExtensionsSchema
from django_ca.constants import ReasonFlags
from django_ca.extensions import serialize_extension
from django_ca.models import Certificate, CertificateAuthority, X509CertMixin
from django_ca.typehints import SerializedExtension


class X509BaseSchema(ModelSchema, abc.ABC):
    """Base schema for CAs and Certificates."""

    created: datetime = Field(description="When the certificate was created.", example=DATETIME_EXAMPLE)
    not_after: datetime = Field(
        description="The certificate is not valid after this date.", example=DATETIME_EXAMPLE
    )
    not_before: datetime = Field(
        description="The certificate is not valid before this date.", example=DATETIME_EXAMPLE
    )
    pem: str = Field(
        description="The public key formatted as PEM.",
        example="-----BEGIN CERTIFICATE-----\n...-----END CERTIFICATE-----\n",
    )
    serial: str = Field(description="Serial (in hex) of the certificate.", example="ABC...0123")
    subject: str = Field(
        description="The subject as RFC 4514 formatted string.",
        example="CN=example.com,O=Example,ST=Vienna,C=AT",
    )
    revoked: bool = Field(description="If the certificate was revoked.", example=False)
    updated: datetime = Field(description="When the certificate was last updated.", example=DATETIME_EXAMPLE)

    class Config:  # pylint: disable=missing-class-docstring
        model = X509CertMixin
        model_fields = sorted(["created", "revoked", "serial", "updated"])

    @staticmethod
    def resolve_created(obj: CertificateAuthority) -> datetime:
        """Strip microseconds from the attribute."""
        return obj.created.replace(microsecond=0)

    @staticmethod
    def resolve_pem(obj: CertificateAuthority) -> str:
        """Convert the public certificate to its PEM format"""
        return obj.pub.pem

    @staticmethod
    def resolve_subject(obj: CertificateAuthority) -> str:
        """Convert the subject to its RFC 4514 representation."""
        return obj.subject.rfc4514_string()

    @staticmethod
    def resolve_updated(obj: CertificateAuthority) -> datetime:
        """Strip microseconds from the attribute."""
        return obj.updated.replace(microsecond=0)


class CertificateAuthorityBaseSchema(ModelSchema, abc.ABC):
    """Base schema for certificate authorities.

    Contains all fields that can be read or updated.
    """

    name: str = Field(description="The human-readable name of the certificate authority.")

    class Config:  # pylint: disable=missing-class-docstring
        model = CertificateAuthority
        model_fields = [
            "name",
            "caa_identity",
            "website",
            "terms_of_service",
            "crl_url",
            "issuer_url",
            "ocsp_url",
            "issuer_alt_name",
            "sign_certificate_policies",
            "ocsp_responder_key_validity",
            "ocsp_response_validity",
            "acme_enabled",
            "acme_registration",
            "acme_profile",
            "acme_requires_contact",
        ]

    @staticmethod
    def resolve_sign_certificate_policies(obj: CertificateAuthority) -> Optional[SerializedExtension]:
        """Convert cryptography extensions to JSON serializable objects."""
        if obj.sign_certificate_policies is None:
            return None
        return serialize_extension(obj.sign_certificate_policies)


class CertificateAuthoritySchema(CertificateAuthorityBaseSchema, X509BaseSchema):
    """Schema for serializing a certificate authority."""

    can_sign_certificates: bool = Field(
        description="If the certificate authority can be used to sign certificates via the API."
    )

    class Config(X509BaseSchema.Config):  # pylint: disable=missing-class-docstring
        model = CertificateAuthority
        model_fields = sorted(
            X509BaseSchema.Config.model_fields + CertificateAuthorityBaseSchema.Config.model_fields
        )

    @staticmethod
    def resolve_can_sign_certificates(obj: CertificateAuthority) -> bool:
        """Strip microseconds from the attribute."""
        return obj.key_exists


class CertificateAuthorityFilterSchema(Schema):
    """Filter-schema for listing certificate authorities."""

    expired: bool = Field(default=False, description="Include expired CAs.")


class CertificateAuthorityUpdateSchema(CertificateAuthorityBaseSchema):
    """Schema for updating certificate authorities."""

    # TYPE NOTE: model_fields_optional does not capture explicitly named fields, so we repeat this
    # with Optional[str], which is an incompatible override
    name: Optional[str] = Field(  # type: ignore[assignment]
        description="The human-readable name of the certificate authority.", required=False
    )

    class Config(CertificateAuthorityBaseSchema.Config):  # pylint: disable=missing-class-docstring
        model_fields_optional = "__all__"


class CertificateSchema(X509BaseSchema):
    """Schema for serializing a certificate."""

    autogenerated: bool = Field(
        description="If the field was automatically generated (e.g. for an OCSP responder)."
    )
    profile: str = Field(description="The profile that the certificate was generated with.")

    class Config(X509BaseSchema.Config):  # pylint: disable=missing-class-docstring
        model = Certificate
        model_fields = sorted(X509BaseSchema.Config.model_fields + ["autogenerated", "profile"])


class CertificateFilterSchema(Schema):
    """Filter schema for certificates."""

    autogenerated: bool = Field(
        default=False, description="Include auto-generated certificates (e.g. OCSP responder certificates)."
    )
    expired: bool = Field(default=False, description="Include expired certificates.")
    profile: Optional[str] = Field(
        description="Only return certificates generated with the given profile.",
        default=None,
        enum=sorted(ca_settings.CA_PROFILES),
    )
    revoked: bool = Field(default=False, description="Include revoked certificates.")


class SignCertificateSchema(Schema):
    """Schema for signing certificates."""

    algorithm: Optional[str] = Field(
        default=None,
        description="Hash algorithm used for signing (default: same as in the certificate authority).",
        enum=sorted(constants.HASH_ALGORITHM_TYPES),
    )
    autogenerated: bool = Field(
        default=False, description="If the certificate should be marked as auto-generated."
    )
    csr: str = Field(
        title="CSR",
        description="The certificate signing request (CSR) in PEM format",
        example="-----BEGIN CERTIFICATE REQUEST-----\n...\n-----END CERTIFICATE REQUEST-----\n",
    )
    expires: Optional[datetime] = Field(
        description="When the certificate is due to expire, defaults to the CA_DEFAULT_EXPIRES setting.",
        example=DATETIME_EXAMPLE,
    )
    extensions: Optional[ExtensionsSchema] = Field(
        default=None, description="**Optional** additional extensions to add to the certificate."
    )
    profile: Optional[str] = Field(
        description="Issue the certificate with the given profile.",
        default=ca_settings.CA_DEFAULT_PROFILE,
        enum=sorted(ca_settings.CA_PROFILES),
    )
    subject: str = Field(
        description="The subject as RFC 4514 formatted string.",
        example="CN=example.com,O=Example,ST=Vienna,C=AT",
    )


class RevokeCertificateSchema(Schema):
    """Schema for revoking certificates."""

    compromised: Optional[datetime] = Field(default=None, description="When the certificate was compromised.")

    reason: ReasonFlags = Field(
        default=ReasonFlags.unspecified,
        description="""The reason why the certificate was revoked. Valid values are `unspecified`,
        `keyCompromise`, `cACompromise`, `affiliationChanged`, `superseeded`, `cessationOfOperation`, 
        `certificateHold`, `privilegeWithdrawn`, `aACompromise` and `removeFromCRL`.""",
    )



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.
#
# Generated by Django 4.1.7 on 2023-03-11 14:11

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0028_alter_certificate_revoked_reason_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="certificate",
            name="updated",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="certificateauthority",
            name="updated",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="certificate",
            name="created",
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name="certificateauthority",
            name="created",
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""
**django-ca** adds a few custom Django signals to important events to let you execute custom actions when
these events happen. Please see `Django's documentation on signals
<https://docs.djangoproject.com/en/dev/ref/signals/>`_ for further information on how to use signals.

If you installed **django-ca** :doc:`from source <quickstart_from_source>`, use the :ref:`CA_CUSTOM_APPS
<settings-ca-custom-apps>` setting to add a custom django app. Please see the `Django documentation on apps
<https://docs.djangoproject.com/en/dev/ref/applications/>`_ if you need help on writing Django apps.
"""

import django.dispatch

pre_create_ca = django.dispatch.Signal()
"""Called before a new certificate authority is created.

Parameters
----------

name : str
    The name of the future CA.
**kwargs
"""

post_create_ca = django.dispatch.Signal()
"""Called after a new certificate authority was created.

Parameters
----------

ca : :py:class:`~django_ca.models.CertificateAuthority`
    The certificate authority that was just created.
"""

post_issue_cert = django.dispatch.Signal()
"""Called after a new certificate was issued.

Parameters
----------

cert : :py:class:`~django_ca.models.Certificate`
    The certificate that was just issued.
"""

pre_sign_cert = django.dispatch.Signal()
"""Called before signing a certificate.

Parameters
----------

ca : :py:class:`~django_ca.models.CertificateAuthority`
    The certificate authority used to sign the certificate
csr : :py:class:`~cg:cryptography.x509.CertificateSigningRequest`
    The certificate signing request used for the certificate.
expires : datetime
    When the certificate will expire.
algorithm : :class:`~cg:cryptography.hazmat.primitives.hashes.HashAlgorithm`
    The algorithm used for signing the certificate.
subject : :class:`~cg:cryptography.x509.Name`
    The subject for the certificate.
extensions : list of :py:class:`~cg:cryptography.x509.Extension`
    The extensions that will be added to the certificate.
password : str or bytes, optional
    The password used for accessing the private key of the certificate authority.
"""


post_sign_cert = django.dispatch.Signal()
"""Called after signing a certificate.

Parameters
----------

ca : :py:class:`~django_ca.models.CertificateAuthority`
    The Certificate Authority used to sign the certificate
cert : :py:class:`~cg:cryptography.x509.Certificate`
    The raw certificate that was just created.
"""

pre_revoke_cert = django.dispatch.Signal()
"""Called before a certificate is revoked.

Parameters
----------

ca
csr
**kwargs
"""

post_revoke_cert = django.dispatch.Signal()
"""Called after a certificate was revoked

Parameters
----------

cert : :py:class:`~django_ca.models.Certificate`
    The certificate that was just revoked.
"""



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to import a certificate.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

import argparse
import typing
from typing import Any

from cryptography import x509

from django.core.management.base import CommandError, CommandParser

from django_ca.management.base import BaseCommand
from django_ca.models import Certificate, CertificateAuthority


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = """Import an existing certificate.

The authority that that signed the certificate must exist in the database."""

    def add_arguments(self, parser: CommandParser) -> None:
        self.add_ca(parser, allow_disabled=False)
        parser.add_argument(
            "pub", help="Path to the public key (PEM or DER format).", type=argparse.FileType("rb")
        )

    def handle(self, pub: typing.BinaryIO, ca: CertificateAuthority, **options: Any) -> None:
        pub_data = pub.read()

        # close reader objects (otherwise we get a ResourceWarning)
        pub.close()

        # load public key
        try:
            pub_loaded = x509.load_pem_x509_certificate(pub_data)
        except Exception:  # pylint: disable=broad-except
            try:
                pub_loaded = x509.load_der_x509_certificate(pub_data)
            except Exception as ex:
                raise CommandError("Unable to load public key.") from ex

        cert = Certificate(ca=ca)
        cert.update_certificate(pub_loaded)
        cert.save()



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Specialized Django forms for the admin interface."""

import typing
from datetime import date, datetime
from typing import Any, Dict, Iterable, List, Optional, Tuple

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.x509.oid import NameOID

from django import forms
from django.contrib.admin.widgets import AdminDateWidget, AdminSplitDateTime
from django.forms.models import ModelFormOptions
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from django_ca import ca_settings, constants, fields
from django_ca.models import Certificate, CertificateAuthority, X509CertMixin
from django_ca.querysets import CertificateAuthorityQuerySet
from django_ca.utils import parse_general_name
from django_ca.widgets import ProfileWidget

if typing.TYPE_CHECKING:
    CertificateModelForm = forms.ModelForm[Certificate]
    X509CertMixinModelForm = forms.ModelForm[X509CertMixin]
else:
    CertificateModelForm = X509CertMixinModelForm = forms.ModelForm

HASH_ALGORITHM_CHOICES = tuple(
    [("", "None")] + sorted([(name, name) for name in constants.HASH_ALGORITHM_TYPES], key=lambda t: t[1])
)


def _initial_expires() -> datetime:
    return datetime.today() + ca_settings.CA_DEFAULT_EXPIRES


def _profile_choices() -> Iterable[Tuple[str, str]]:
    return sorted([(p, p) for p in ca_settings.CA_PROFILES], key=lambda e: e[0])


class X509CertMixinAdminForm(X509CertMixinModelForm):
    """Admin form to add a dynamic help text to the ``pub`` field.

    The help_text is set by adding a value to the help_texts dictionary of the models Meta class. We use this
    unusual way because it should contain links referencing the currently displayed object and the normal
    methods do not work this way:

    * You cannot use the normal way of setting ``help_texts`` in the forms ``Meta`` class, because we cannot
      reference the object instance here.
    * We cannot access self.fields['pub'] in the constructor, because it is a readonly field and thus not
      present in the form.
    """

    _meta: "ModelFormOptions[X509CertMixin]"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        meta = self._meta
        if meta.help_texts is None:  # pragma: no cover
            # help_texts is always set since we have a Meta class, but keeping this here as a precaution.
            meta.help_texts = {}

        info = f"{self.instance._meta.app_label}_{self.instance._meta.model_name}"
        url = reverse(f"admin:{info}_download", kwargs={"pk": self.instance.pk})
        bundle_url = reverse(f"admin:{info}_download_bundle", kwargs={"pk": self.instance.pk})
        meta.help_texts["pub_pem"] = _(
            'Download: <a href="%s?format=PEM">as PEM</a> | <a href="%s?format=DER">as DER</a><br />'
            'Certificate bundle: <a href="%s?format=PEM">as PEM</a>'
        ) % (url, url, bundle_url)


class CertificateAuthorityForm(X509CertMixinAdminForm):
    """Admin form for :py:class:`django_ca.models.CertificateAuthority`."""

    acme_profile = forms.ChoiceField(
        label=_("Profile"),
        required=True,
        widget=ProfileWidget,
        help_text=_("Profile used when issuing certificates via ACMEv2."),
        initial=ca_settings.CA_DEFAULT_PROFILE,
        choices=_profile_choices,
    )
    sign_certificate_policies = fields.CertificatePoliciesField(required=False)

    class Meta:
        labels = {"acme_registration": _("Account registration")}


class CreateCertificateBaseForm(CertificateModelForm):
    """Base class for forms that create a certificate.

    This is used by forms for creating a new certificate and resigning an existing one."""

    password = forms.CharField(
        widget=forms.PasswordInput,
        required=False,
        help_text=_("Password for the private key. If not given, the private key must be unencrypted."),
    )
    expires = forms.DateField(initial=_initial_expires, widget=AdminDateWidget())
    subject = fields.SubjectField(label=_("Subject"), required=False)
    subject_alternative_name = fields.SubjectAlternativeNameField(
        required=False,
        help_text=_("""Coma-separated list of alternative names for the certificate."""),
    )
    profile = forms.ChoiceField(
        required=False,
        widget=ProfileWidget,
        help_text=_("Select a suitable profile or manually select X509 extensions below."),
        initial=ca_settings.CA_DEFAULT_PROFILE,
        choices=_profile_choices,
    )
    algorithm = forms.ChoiceField(
        required=False,
        label=_("Signature hash algorithm"),
        initial=constants.HASH_ALGORITHM_NAMES[type(ca_settings.CA_DEFAULT_SIGNATURE_HASH_ALGORITHM)],
        choices=HASH_ALGORITHM_CHOICES,
        help_text=_(
            "SHA-512 is fine for RSA/EC-based certificate authorities, choose None for Ed448/Ed25519-based "
            "certificate authorities and SHA-256 for DSA-based certificate authorities."
        ),
    )
    autogenerated = forms.BooleanField(
        required=False, help_text=_("If this certificate was automatically generated.")
    )

    authority_information_access = fields.AuthorityInformationAccessField(
        help_text=_("Indicates how to access information and services of the issuer.")
    )
    certificate_policies = fields.CertificatePoliciesField(required=False)
    crl_distribution_points = fields.CRLDistributionPointField(
        help_text=_("Where to retrieve Certificate Revocation Lists."),
    )
    freshest_crl = fields.FreshestCRLField(
        help_text=_("Where to retrieve Delta CRLs (rarely used, not supported by django-ca).")
    )
    issuer_alternative_name = fields.IssuerAlternativeNameField(
        help_text=_("Additional information about the issuer (rarely used).")
    )
    key_usage = fields.KeyUsageField(help_text=_("Defines the purpose of the certificate."))
    extended_key_usage = fields.ExtendedKeyUsageField(
        help_text=_("Defines additional purposes of the certificate.")
    )
    ocsp_no_check = fields.OCSPNoCheckField(
        help_text=_(
            "Indicates that an OCSP client can trust the certificate. Used in OCSP responder certificates."
        )
    )
    tls_feature = fields.TLSFeatureField(
        help_text=_("Allows enforcing TLS protocol features. Only status_request is commonly used.")
    )

    # Prevent auto-completing the password field. Browsers will otherwise prefill this field with the *users*
    # password, which is usually the wrong password. Especially annoying for CAs without a password, as the
    # browser will prevent form submission without entering a different non-empty password.
    password.widget.attrs.update({"autocomplete": "new-password"})

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        # Set choices, so we can filter out CAs where the private key does not exist locally
        field = typing.cast(forms.ModelChoiceField, self.fields["ca"])
        qs = typing.cast(CertificateAuthorityQuerySet, field.queryset)
        field.choices = [
            (field.prepare_value(ca), field.label_from_instance(ca))
            for ca in qs.filter(enabled=True)
            if ca.key_exists
        ]

    def clean_algorithm(self) -> Optional[hashes.HashAlgorithm]:  # pylint: disable=missing-function-docstring
        if algorithm_name := self.cleaned_data["algorithm"]:
            return constants.HASH_ALGORITHM_TYPES[algorithm_name]()
        return None  # required by mypy

    def clean_expires(self) -> datetime:  # pylint: disable=missing-function-docstring
        expires: datetime = self.cleaned_data["expires"]
        if expires < date.today():
            raise forms.ValidationError(_("Certificate cannot expire in the past."))
        return expires

    def clean_password(self) -> Optional[bytes]:  # pylint: disable=missing-function-docstring
        password: str = self.cleaned_data["password"]
        if not password:
            return None
        return password.encode("utf-8")

    def clean(self) -> Optional[Dict[str, Any]]:
        data = super().clean()

        # COVERAGE Unclear if/when this happens, but django-stubs==1.15.0 reports data as Optional.
        if data is None:  # pragma: no cover
            return data

        expires = data.get("expires")
        ca: CertificateAuthority = data["ca"]
        password = typing.cast(Optional[str], data.get("password"))
        subject = typing.cast(Optional[x509.Name], data.get("subject"))
        algorithm = typing.cast(Optional[hashes.HashAlgorithm], data.get("algorithm"))
        subject_alternative_name, include_common_name = data.get("subject_alternative_name", (None, False))

        subject_alternative_name = typing.cast(
            Optional[x509.Extension[x509.SubjectAlternativeName]], subject_alternative_name
        )

        # Load the CA to test the password
        try:
            ca.key(password)
        except Exception as e:  # pylint: disable=broad-except; for simplicity
            self.add_error("password", str(e))

        if ca.key_type in ("Ed448", "Ed25519") and algorithm is not None:
            self.add_error(
                "algorithm",
                _("%(key_type)s-based certificate authorities do not use a signature hash algorithm.")
                % {"key_type": ca.key_type},
            )
        elif ca.key_type == "DSA" and not isinstance(algorithm, hashes.SHA256):
            self.add_error(
                "algorithm",
                _("DSA-based certificate authorities require a SHA-256 signature hash algorithm."),
            )
        elif ca.key_type not in ("Ed448", "Ed25519") and algorithm is None:
            self.add_error(
                "algorithm",
                _("%(key_type)s-based certificate authorities require a signature hash algorithm.")
                % {"key_type": ca.key_type},
            )

        common_names: List[x509.NameAttribute] = []
        if subject is not None:
            common_names = subject.get_attributes_for_oid(NameOID.COMMON_NAME)

        # If the user decided to include any Common Names in the Subject Alternative Name extension, we check
        # that they can be parsed as general name.
        if include_common_name:
            for common_name in common_names:
                try:
                    parse_general_name(common_name.value)  # type: ignore[arg-type]
                except ValueError:
                    self.add_error(
                        "subject_alternative_name",
                        _(
                            "The CommonName cannot be parsed as general name. Either change the "
                            "CommonName or do not include it."
                        ),
                    )

        # Make sure that we have at least a Common Name *or* a Subject Alternative Name extension.
        if subject is not None and not common_names and not subject_alternative_name:
            self.add_error(
                "subject_alternative_name",
                "Subject Alternative Name is required if the subject does not contain a Common Name.",
            )

        if ca and expires and ca.expires.date() < expires:
            stamp = ca.expires.strftime("%Y-%m-%d")
            self.add_error("expires", _("CA expires on %s, certificate must not expire after that.") % stamp)
        return data

    class Media:
        js = (
            "admin/js/jquery.init.js",
            "django_ca/admin/js/extensions.js",
            "django_ca/admin/js/ca-details.js",
        )

    class Meta:
        model = Certificate
        fields = [
            "watchers",
            "ca",
        ]


class CreateCertificateForm(CreateCertificateBaseForm):
    """Admin form for creating a completely new certificate."""

    class Meta:
        model = Certificate
        fields = [
            "csr",
            "watchers",
            "ca",
        ]


class ResignCertificateForm(CreateCertificateBaseForm):
    """Admin form for resigning an existing certificate."""


class RevokeCertificateForm(CertificateModelForm):
    """Admin form for revoking a certificate."""

    class Media:
        js = (
            # jquery/core.js for the datetime widgets:
            "admin/js/vendor/jquery/jquery.js",
            "admin/js/jquery.init.js",
            "admin/js/core.js",
        )

    class Meta:
        model = Certificate
        fields = ["revoked_reason", "compromised"]
        field_classes = {
            "compromised": forms.SplitDateTimeField,
        }
        widgets = {
            "compromised": AdminSplitDateTime,
        }



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to cache CRLs.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

import argparse
from typing import Any, List

from django_ca.management.base import BaseCommand
from django_ca.tasks import cache_crls, run_task


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = "Cache CRLs"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "serial",
            nargs="*",
            help="Generate CRLs for the given CAs. If omitted, generate CRLs for all CAs.",
        )

    def handle(self, serial: List[str], **options: Any) -> None:
        run_task(cache_crls, serial)



# Generated by Django 4.2.3 on 2023-07-31 17:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0033_alter_certificate_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="certificateauthority",
            name="acme_registration",
            field=models.BooleanField(
                default=True,
                help_text="Allow ACME clients to register new accounts.",
                verbose_name="ACME account registration.",
            ),
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""``django_ca.extensions.serialize`` contains functions to serialize extensions."""

import binascii
from typing import Any, Dict, List, Optional

from cryptography import x509
from cryptography.x509.oid import AuthorityInformationAccessOID

from django_ca import typehints
from django_ca.constants import EXTENDED_KEY_USAGE_NAMES, KEY_USAGE_NAMES, LOG_ENTRY_TYPE_KEYS
from django_ca.typehints import (
    PolicyQualifier,
    SerializedExtension,
    SerializedPolicyInformation,
    SerializedPolicyQualifier,
    SerializedPolicyQualifiers,
    SerializedUserNotice,
)
from django_ca.utils import bytes_to_hex, format_general_name, format_name


def _authority_information_access_serialized(
    value: typehints.InformationAccessExtensionType,
) -> Dict[str, List[str]]:
    descriptions = {}
    issuers = [ad for ad in value if ad.access_method == AuthorityInformationAccessOID.CA_ISSUERS]
    ocsp = [ad for ad in value if ad.access_method == AuthorityInformationAccessOID.OCSP]
    if issuers:
        descriptions["issuers"] = [format_general_name(ad.access_location) for ad in issuers]
    if ocsp:
        descriptions["ocsp"] = [format_general_name(ad.access_location) for ad in ocsp]
    return descriptions


def _authority_key_identifier_serialized(
    value: x509.AuthorityKeyIdentifier,
) -> typehints.SerializedAuthorityKeyIdentifier:
    serialized: typehints.SerializedAuthorityKeyIdentifier = {}
    if value.key_identifier:
        serialized["key_identifier"] = bytes_to_hex(value.key_identifier)
    if value.authority_cert_serial_number is not None:
        serialized["authority_cert_serial_number"] = value.authority_cert_serial_number
    if value.authority_cert_issuer:
        serialized["authority_cert_issuer"] = [
            format_general_name(aci) for aci in value.authority_cert_issuer
        ]
    return serialized


def _basic_constraints_serialized(value: x509.BasicConstraints) -> typehints.SerializedBasicConstraints:
    serialized: typehints.SerializedBasicConstraints = {"ca": value.ca}
    if value.ca is True:
        serialized["path_length"] = value.path_length
    return serialized


def _serialize_policy_qualifier(qualifier: PolicyQualifier) -> SerializedPolicyQualifier:
    if isinstance(qualifier, str):
        return qualifier

    value: SerializedUserNotice = {}
    if qualifier.explicit_text:
        value["explicit_text"] = qualifier.explicit_text

    if qualifier.notice_reference is not None:
        value["notice_reference"] = {
            "notice_numbers": qualifier.notice_reference.notice_numbers,
        }
        if qualifier.notice_reference.organization is not None:
            value["notice_reference"]["organization"] = qualifier.notice_reference.organization
    return value


def _serialize_policy_information(policy_information: x509.PolicyInformation) -> SerializedPolicyInformation:
    policy_qualifiers: Optional[SerializedPolicyQualifiers] = None
    if policy_information.policy_qualifiers is not None:
        policy_qualifiers = [_serialize_policy_qualifier(q) for q in policy_information.policy_qualifiers]

    serialized: SerializedPolicyInformation = {
        "policy_identifier": policy_information.policy_identifier.dotted_string,
        "policy_qualifiers": policy_qualifiers,
    }
    return serialized


def _certificate_policies_serialized(value: x509.CertificatePolicies) -> List[SerializedPolicyInformation]:
    return [_serialize_policy_information(pi) for pi in value]


def _distribution_points_serialized(
    value: typehints.CRLExtensionType,
) -> List[typehints.SerializedDistributionPoint]:
    points: List[typehints.SerializedDistributionPoint] = []

    for distribution_point in value:
        point: typehints.SerializedDistributionPoint = {}
        if distribution_point.full_name:
            point["full_name"] = [format_general_name(name) for name in distribution_point.full_name]
        elif distribution_point.relative_name:  # pragma: no branch  # Distribution Point has only these two
            point["relative_name"] = format_name(distribution_point.relative_name)

        if distribution_point.crl_issuer:
            point["crl_issuer"] = [format_general_name(name) for name in distribution_point.crl_issuer]
        if distribution_point.reasons:
            point["reasons"] = sorted([r.name for r in distribution_point.reasons])

        points.append(point)
    return points


def _key_usage_serialized(value: x509.KeyUsage) -> List[str]:
    values: List[str] = []
    for attr in KEY_USAGE_NAMES:
        try:
            if getattr(value, attr):
                values.append(attr)
        except ValueError:
            # x509.KeyUsage raises ValueError on some attributes to ensure consistency
            pass
    return sorted(values)


def _name_constraints_serialized(value: x509.NameConstraints) -> typehints.SerializedNameConstraints:
    serialized: typehints.SerializedNameConstraints = {}
    if value.permitted_subtrees:
        serialized["permitted"] = [format_general_name(name) for name in value.permitted_subtrees]
    if value.excluded_subtrees:
        serialized["excluded"] = [format_general_name(name) for name in value.excluded_subtrees]
    return serialized


def _policy_constraints_serialized(value: x509.PolicyConstraints) -> typehints.SerializedPolicyConstraints:
    serialized: typehints.SerializedPolicyConstraints = {}
    if value.inhibit_policy_mapping is not None:
        serialized["inhibit_policy_mapping"] = value.inhibit_policy_mapping
    if value.require_explicit_policy is not None:
        serialized["require_explicit_policy"] = value.require_explicit_policy
    return serialized


def _signed_certificate_timestamps_serialized(
    value: typehints.SignedCertificateTimestampType,
) -> List[typehints.SerializedSignedCertificateTimestamp]:
    time_format = "%Y-%m-%d %H:%M:%S.%f"
    return [
        {
            "log_id": binascii.hexlify(sct.log_id).decode("utf-8"),
            "timestamp": sct.timestamp.strftime(time_format),
            "type": LOG_ENTRY_TYPE_KEYS[sct.entry_type],
            "version": sct.version.name,
        }
        for sct in value
    ]


def _tls_feature_serialized(value: x509.TLSFeature) -> List[str]:
    serialized: List[str] = [feature.name for feature in value]
    return sorted(serialized)


def _serialize_extension(  # pylint: disable=too-many-return-statements
    value: x509.ExtensionType,
) -> Any:
    if isinstance(value, (x509.OCSPNoCheck, x509.PrecertPoison)):
        return None
    if isinstance(value, (x509.IssuerAlternativeName, x509.SubjectAlternativeName)):
        return [format_general_name(name) for name in value]
    if isinstance(value, (x509.AuthorityInformationAccess, x509.SubjectInformationAccess)):
        return _authority_information_access_serialized(value)
    if isinstance(value, (x509.FreshestCRL, x509.CRLDistributionPoints)):
        return _distribution_points_serialized(value)
    if isinstance(value, (x509.PrecertificateSignedCertificateTimestamps, x509.SignedCertificateTimestamps)):
        return _signed_certificate_timestamps_serialized(value)
    if isinstance(value, x509.AuthorityKeyIdentifier):
        return _authority_key_identifier_serialized(value)
    if isinstance(value, x509.BasicConstraints):
        return _basic_constraints_serialized(value)
    if isinstance(value, x509.CertificatePolicies):
        return _certificate_policies_serialized(value)
    if isinstance(value, x509.ExtendedKeyUsage):
        return sorted([EXTENDED_KEY_USAGE_NAMES[usage] for usage in value])
    if isinstance(value, x509.InhibitAnyPolicy):
        return value.skip_certs
    if isinstance(value, x509.KeyUsage):
        return _key_usage_serialized(value)
    if isinstance(value, x509.NameConstraints):
        return _name_constraints_serialized(value)
    if isinstance(value, x509.PolicyConstraints):
        return _policy_constraints_serialized(value)
    if isinstance(value, x509.SubjectKeyIdentifier):
        return bytes_to_hex(value.key_identifier)
    if isinstance(value, x509.TLSFeature):
        return _tls_feature_serialized(value)
    if isinstance(value, x509.UnrecognizedExtension):
        return bytes_to_hex(value.value)
    if isinstance(value, x509.ExtensionType):
        raise TypeError(
            f"{value.__class__.__name__} (oid: {value.oid.dotted_string}): Unknown extension type."
        )
    raise TypeError(f"{value.__class__.__name__}: Not a cryptography.x509.ExtensionType.")


def serialize_extension(extension: x509.Extension[x509.ExtensionType]) -> SerializedExtension:
    """Serialize an extension to a dictionary.

    This is the inverse of :py:func:`~django_ca.extensions.parse_extension` and is used to serialize
    extension information for API calls in the admin interface.
    """

    value = _serialize_extension(extension.value)
    serialized: SerializedExtension = {"critical": extension.critical, "value": value}
    return serialized



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""API utility functions."""
import typing
from typing import Any

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.http import Http404

from django_ca.models import Certificate, CertificateAuthority

if typing.TYPE_CHECKING:
    from django.contrib.auth.models import User  # pylint: disable=imported-auth-user
else:
    from django.contrib.auth import get_user_model

    User = get_user_model()


def get_certificate_authority(serial: str, expired: bool = False) -> CertificateAuthority:
    """Get a certificate authority from the given serial."""
    qs = CertificateAuthority.objects.enabled()
    if expired is False:
        qs = qs.valid()

    try:
        return qs.get(serial=serial)
    except CertificateAuthority.DoesNotExist as ex:
        raise Http404(f"{serial}: Certificate authority not found.") from ex


@transaction.atomic
def create_api_user(
    username: str,
    password: str,
    view_certificateauthority: bool = True,
    change_certificateauthority: bool = True,
    sign_certificate: bool = True,
    view_certificate: bool = True,
    revoke_certificate: bool = True,
    **extra_fields: Any,
) -> User:
    """Create an API user capable of using the REST API.

    By default, the user will be able to perform all actions provided by the API.

    Note that *unlike* :py:meth:`~django:django.contrib.auth.models.UserManager.create_user`, the `password`
    argument is the second argument and mandatory. You can still pass an `email` address as keyword argument.

    >>> create_api_user("username", "password", revoke_certificate=False, email="user@example.com")

    Parameters
    ----------

    username : str
        The username for the API user.
    password : str
        The password for the API user.
    view_certificateauthority : bool, optional
        If the user is able to list/view certificate authorities via the API.
    change_certificateauthority : bool, optional
        If the user is able to update certificate authorities via the API.
    sign_certificate : bool, optional
        If the user is able to sign new certificates via the API.
    view_certificate : bool, optional
        If the user is able to list/view existing certificates via the API.
    revoke_certificate : bool, optional
        If the user is able to revoke certificates via the API.
    **extra_fields
        Any additional keyword arguments are passed to
        :py:meth:`~django:django.contrib.auth.models.UserManager.create_user`.
    """
    ca_content_type = ContentType.objects.get_for_model(CertificateAuthority)
    cert_content_type = ContentType.objects.get_for_model(Certificate)

    user = User.objects.create_user(username, password=password, **extra_fields)

    permissions = []
    if view_certificateauthority is True:
        permissions.append(
            Permission.objects.get(codename="view_certificateauthority", content_type=ca_content_type)
        )
    if change_certificateauthority is True:
        permissions.append(
            Permission.objects.get(codename="change_certificateauthority", content_type=ca_content_type)
        )
    if sign_certificate is True:
        permissions.append(
            Permission.objects.get(codename="sign_certificate", content_type=cert_content_type)
        )
    if view_certificate is True:
        permissions.append(
            Permission.objects.get(codename="view_certificate", content_type=cert_content_type)
        )
    if revoke_certificate is True:
        permissions.append(
            Permission.objects.get(codename="revoke_certificate", content_type=cert_content_type)
        )
    user.user_permissions.add(*permissions)

    return user
